// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <rcsc/soccer_math.h>
#include <rcsc/param/server_param.h>

#include <rcsc/player/logger.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>

#include "bhv_defender_get_ball.h"

/*-------------------------------------------------------------------*/
/*!
  execute action
*/
bool
Bhv_DefenderGetBall::execute( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: Bhv_DefenderGetBall"
                        ,__FILE__, __LINE__ );

    const rcsc::WorldModel & wm = agent->world();
    //--------------------------------------------------
    // intercept
    int self_min = wm.getInterceptTable()->getSelfReachCycle();
    int mate_min = wm.getInterceptTable()->getTeammateReachCycle();
    int opp_min = wm.getInterceptTable()->getOpponentReachCycle();

    //////////////////////////////////////////////////////////////////
    // intercept normal
    if ( self_min < mate_min && self_min < opp_min - 3 )
    {
        agent->debugClient().addMessage( "DefGetBall(1)" );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: get ball normal"
                            ,__FILE__, __LINE__ );
        rcsc::Body_Intercept().execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToBall() );
        return true;
    }

    //////////////////////////////////////////////////////////////////
    // attack to ball owner
    const rcsc::PlayerObject * ball_nearest = wm.getTeammateNearestToBall( 10 );
    if ( ( ball_nearest
           && ( ball_nearest->distFromBall() > wm.ball().distFromSelf()
                || wm.ball().distFromSelf() < 5.0 ) )
         && wm.self().pos().x < M_home_pos.x + 5.0
         && ( wm.existKickableOpponent() || opp_min < 3 )
         )
    {
        if ( wm.existKickableOpponent() )
        {
            rcsc::Vector2D opp_pos = wm.getOpponentsFromBall().front()->pos();
            opp_pos.x -= 0.2;
            agent->debugClient().setTarget( opp_pos );
            agent->debugClient().addMessage( "DefGetBall(2)" );
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: attack ball owner (%f, %f)"
                                ,__FILE__, __LINE__,
                                opp_pos.x, opp_pos.y );
            rcsc::Body_GoToPoint( opp_pos,
                                  0.1,
                                  rcsc::ServerParam::i().maxPower()
                                  ).execute( agent );
            agent->setNeckAction( new rcsc::Neck_TurnToBall() );
        }
        else
        {
            rcsc::Vector2D ball_future
                = rcsc::inertia_n_step_point( wm.ball().pos(),
                                              wm.ball().vel(),
                                              opp_min,
                                              rcsc::ServerParam::i().ballDecay() );
            ball_future.x -= 0.2;
            agent->debugClient().setTarget( ball_future );
            agent->debugClient().addMessage( "DefGetBall(3)" );
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: attack ball future (%f, %f)"
                                ,__FILE__, __LINE__,
                                ball_future.x, ball_future.y );
            rcsc::Body_GoToPoint( ball_future,
                                  0.1,
                                  rcsc::ServerParam::i().maxPower()
                                  ).execute( agent );
            agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        }
        return true;
    }

    //////////////////////////////////////////////////////////////////
    // intercept close
    if ( ! wm.existKickableTeammate()
         && ( self_min <= 3
              || wm.getInterceptTable()->isSelfFastestPlayer() )

         )
    {
        agent->debugClient().addMessage( "DefGetBall(4)" );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: get ball close"
                            ,__FILE__, __LINE__ );
        rcsc::Body_Intercept().execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        return true;
    }

    //////////////////////////////////////////////////////////////////
    // attack to ball owner
    if ( ! wm.existKickableOpponent() && opp_min < self_min
         && wm.self().pos().x < M_home_pos.x + 5.0
         )
    {
        rcsc::Vector2D ball_future_pos
            = rcsc::inertia_n_step_point( wm.ball().pos(),
                                          wm.ball().vel(),
                                          opp_min,
                                          rcsc::ServerParam::i().ballDecay() );
        ball_future_pos.x -= 0.5;
        double my_dist = wm.self().pos().dist(ball_future_pos);
        if ( my_dist < 5.0 )
        {
            double mate_dist = 1000.0;
            const rcsc::PlayerObject * close_mate
                = wm.getTeammateNearestTo( ball_future_pos, 10, &mate_dist );
            if ( ! close_mate || mate_dist > my_dist )
            {
                agent->debugClient().addMessage( "DefGetBall(5)" );
                agent->debugClient().setTarget( ball_future_pos );
                rcsc::dlog.addText( rcsc::Logger::TEAM,
                                    "%s:%d: attack future ball owner (%f, %f)"
                                    ,__FILE__, __LINE__,
                                    ball_future_pos.x, ball_future_pos.y );
                rcsc::Body_GoToPoint( ball_future_pos,
                                      0.1,
                                      rcsc::ServerParam::i().maxPower()
                                      ).execute( agent );
                agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
                return true;
            }
        }

    }

    rcsc::dlog.addText( rcsc::Logger::TEAM,
                  "%s:%d: failed"
                  ,__FILE__, __LINE__ );
    return false;
}
