// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <rcsc/param/server_param.h>

#include <rcsc/player/logger.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include "bhv_basic_move.h"
#include "bhv_basic_tackle.h"
#include "bhv_mark_close_opponent.h"

#include "bhv_press.h"

/*-------------------------------------------------------------------*/
/*!
  execute action
*/
bool
Bhv_Press::execute( rcsc::PlayerAgent * agent )
{
    static bool S_recover_mode = false;

    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: Bhv_Press"
                        ,__FILE__, __LINE__ );

    const rcsc::WorldModel & wm = agent->world();

    if ( wm.getOpponentsFromBall().empty() )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: no opponent. basic move"
                            ,__FILE__, __LINE__ );
        return Bhv_BasicMove( M_home_pos ).execute( agent );
    }

    ///////////////////////////////////
    // tackle
    if ( Bhv_BasicTackle( 0.85, 90.0 ).execute( agent ) )
    {
        return true;
    }

    ///////////////////////////////////

    int self_min = wm.getInterceptTable()->getSelfReachCycle();
    int opp_min = wm.getInterceptTable()->getOpponentReachCycle();

    if ( ! wm.existKickableTeammate()
         && ( self_min < opp_min + 3 || self_min < 4 )
         )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: get ball"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "Press:get" );
        rcsc::Body_Intercept().execute( agent );
        if ( wm.existKickableOpponent() )
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBall() );
        }
        else
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        }
        return true;
    }

    ///////////////////////////////////
    // set target point

    rcsc::Vector2D base_target = wm.getOpponentsFromBall().front()->pos();

    if ( std::fabs( base_target.y - M_home_pos.y ) > 10.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: y diff is too big. target=(%f, %f), myhome=(%f, %f)"
                            ,__FILE__, __LINE__,
                            base_target.x, base_target.y,
                            M_home_pos.x, M_home_pos.y );
        Bhv_MarkCloseOpponent( M_home_pos ).execute( agent );
        return true;
    }

    rcsc::Vector2D target_point = base_target;



    ////////////////////////////////////////////////////////
    // check teammate closer than self
    const double x_thr = wm.getDefenseLineX() + 5.0;
    double my_dist2 = wm.self().pos().dist2(target_point);

    const rcsc::PlayerPtrCont::const_iterator mates_end
        = wm.getTeammatesFromSelf().end();
    for ( rcsc::PlayerPtrCont::const_iterator it
              = wm.getTeammatesFromSelf().begin();
          it != mates_end;
          ++it )
    {
        if ( (*it)->pos().x < x_thr )
        {
            continue;
        }
        double d2 = (*it)->pos().dist2(target_point);
        if ( d2 < my_dist2 || d2 < 3.0 * 3.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: exist other pressor(%f, %f)"
                                ,__FILE__, __LINE__,
                                (*it)->pos().x, (*it)->pos().y );
            Bhv_BasicMove( M_home_pos ).execute( agent );
            return true;
        }
    }

    ///////////////////////////////////
    if ( target_point.x < wm.self().pos().x + 1.0 )
    {
        target_point.y = wm.self().pos().y;
    }


    ///////////////////////////////////
    // set dash power
    if ( wm.self().stamina() < rcsc::ServerParam::i().staminaMax() * 0.48 )
    {
        S_recover_mode = true;
    }
    if ( wm.self().stamina() > rcsc::ServerParam::i().staminaMax() * 0.75 )
    {
        S_recover_mode = false;
    }

    double dash_power;
    if ( S_recover_mode )
    {
        dash_power = wm.self().playerType().staminaIncMax();
        dash_power *= wm.self().recovery();
        dash_power *= 0.8;
    }
    else if ( wm.ball().pos().x < wm.self().pos().x - 1.0 )
    {
        dash_power = rcsc::ServerParam::i().maxPower() * 0.9;
    }
    else if ( wm.ball().distFromSelf() > 10.0 )
    {
        dash_power = rcsc::ServerParam::i().maxPower() * 0.7;
    }
    else if ( wm.ball().distFromSelf() > 2.0 )
    {
        dash_power = rcsc::ServerParam::i().maxPower() * 0.75;
    }
    else
    {
        dash_power = rcsc::ServerParam::i().maxPower();
    }


    agent->debugClient().addMessage( "press%.0f", dash_power );
    agent->debugClient().setTarget( target_point );

    double dist_thr = wm.ball().distFromSelf() * 0.1;
    if ( dist_thr < 0.5 ) dist_thr = 0.5;
    rcsc::Body_GoToPoint( target_point, dist_thr, dash_power
                          ).execute( agent );

    if ( wm.existKickableOpponent() )
    {
        agent->setNeckAction( new rcsc::Neck_TurnToBall() );
    }
    else
    {
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
    }

    return true;
}
